@@ -31,10 +31,19 @@
 #include "wavpack.h"
 #include "utils.h"
 
+#ifdef _WIN32
+#include "win32_unicode_support.h"
+#define fprintf fprintf_utf8
+#define fputs fputs_utf8
+#define remove(f) unlink_utf8(f)
+#endif
+
 #ifdef _WIN32
 
 int copy_timestamp (const char *src_filename, const char *dst_filename)
 {
+    wchar_t *src_filename_utf16 = utf8_to_utf16(src_filename);
+    wchar_t *dst_filename_utf16 = utf8_to_utf16(dst_filename);
     FILETIME last_modified;
     HANDLE src, dst;
     int res = TRUE;
@@ -42,10 +51,13 @@ int copy_timestamp (const char *src_filename, const char *dst_filename)
     if (*src_filename == '-' || *dst_filename == '-')
         return res;
 
-    src = CreateFile (src_filename, GENERIC_READ, FILE_SHARE_READ, NULL,
+    if (!src_filename_utf16 || !dst_filename_utf16)
+        return FALSE;
+
+    src = CreateFileW (src_filename_utf16, GENERIC_READ, FILE_SHARE_READ, NULL,
          OPEN_EXISTING, FILE_FLAG_SEQUENTIAL_SCAN, NULL);
 
-    dst = CreateFile (dst_filename, GENERIC_WRITE, FILE_SHARE_WRITE, NULL,
+    dst = CreateFileW (dst_filename_utf16, GENERIC_WRITE, FILE_SHARE_WRITE, NULL,
          OPEN_EXISTING, FILE_FLAG_SEQUENTIAL_SCAN, NULL);
 
     if (src == INVALID_HANDLE_VALUE || dst == INVALID_HANDLE_VALUE ||
@@ -59,6 +71,9 @@ int copy_timestamp (const char *src_filename, const char *dst_filename)
     if (dst != INVALID_HANDLE_VALUE)
         CloseHandle (dst);
 
+    free (src_filename_utf16);
+    free (dst_filename_utf16);
+
     return res;
 }
 
@@ -100,18 +115,12 @@ int copy_timestamp(const char *src_filename, const char *dst_filename)
 
 #if defined(_WIN32)
 
-static int is_second_byte (char *filespec, char *pos);
-
 char *filespec_ext (char *filespec)
 {
     char *cp = filespec + strlen (filespec);
-    LANGID langid = GetSystemDefaultLangID ();
 
     while (--cp >= filespec) {
 
-        if (langid == 0x411 && is_second_byte (filespec, cp))
-            --cp;
-
         if (*cp == '\\' || *cp == ':')
             return NULL;
 
@@ -197,8 +206,8 @@ char *filespec_path (char *filespec)
 char *filespec_path (char *filespec)
 {
     char *cp = filespec + strlen (filespec);
-    LANGID langid = GetSystemDefaultLangID ();
-    struct _finddata_t finddata;
+    struct _wfinddata_t wfinddata;
+    wchar_t *filespec_utf16;
     intptr_t file;
 
     if (cp == filespec || filespec_wild (filespec))
@@ -206,23 +215,28 @@ char *filespec_path (char *filespec)
 
     --cp;
 
-    if (langid == 0x411 && is_second_byte (filespec, cp))
-        --cp;
-
     if (*cp == '\\' || *cp == ':')
         return filespec;
 
     if (*cp == '.' && cp == filespec)
         return strcat (filespec, "\\");
 
-    if ((file = _findfirst (filespec, &finddata)) != (intptr_t) -1 &&
-        (finddata.attrib & _A_SUBDIR)) {
+    filespec_utf16 = utf8_to_utf16(filespec);
+
+    if (!filespec_utf16)
+        return NULL;
+
+    if ((file = _wfindfirst (filespec_utf16, &wfinddata)) != (intptr_t) -1 &&
+        (wfinddata.attrib & _A_SUBDIR)) {
             _findclose (file);
+            free (filespec_utf16);
             return strcat (filespec, "\\");
     }
-    if (file != -1L)
-            _findclose(file);
 
+    if (file != -1L)
+        _findclose(file);
+
+    free (filespec_utf16);
     return NULL;
 }
 
@@ -248,11 +262,8 @@ char *filespec_wild (char *filespec)
 char *filespec_name (char *filespec)
 {
     char *cp = filespec + strlen (filespec);
-    LANGID langid = GetSystemDefaultLangID ();
 
     while (--cp >= filespec) {
-        if (langid == 0x411 && is_second_byte (filespec, cp))
-            --cp;
 
         if (*cp == '\\' || *cp == ':')
             break;
@@ -282,27 +293,6 @@ char *filespec_name (char *filespec)
 
 #endif
 
-//////////////////////////////////////////////////////////////////////////////
-// This function returns TRUE if "pos" is pointing to the second byte of a  //
-// double-byte character in the string "filespec" which is assumed to be    //
-// shift-JIS.                                                               //
-//////////////////////////////////////////////////////////////////////////////
-
-#if defined(_WIN32)
-
-static int is_second_byte (char *filespec, char *pos)
-{
-    unsigned char *cp = pos;
-
-    while (cp > (unsigned char *)filespec && ((cp [-1] >= 0x81 && cp [-1] <= 0x9f) ||
-                             (cp [-1] >= 0xe0 && cp [-1] <= 0xfc)))
-        cp--;
-
-    return (int)((unsigned char *)pos - cp) & 1;
-}
-
-#endif
-
 //////////////////////////////////////////////////////////////////////////////
 // This function allows the user to type 'y', 'n', or 'a' (with Enter) in   //
 // response to a system query. The return value is the key typed as         //
@@ -497,32 +487,6 @@ void error_line (char *error, ...)
 
 #endif
 
-void debug_line (char *error, ...)
-{
-    char error_msg [512];
-    va_list argptr;
-
-    if (!debug_logging_mode)
-        return;
-
-    error_msg [0] = '\r';
-    va_start (argptr, error);
-    vsprintf (error_msg + 1, error, argptr);
-    va_end (argptr);
-    fputs (error_msg, stderr);
-    finish_line ();
-
-    if (debug_logging_mode) {
-        FILE *error_log = fopen ("c:\\wavpack.log", "a+");
-
-        if (error_log) {
-            fputs (error_msg + 1, error_log);
-            fputc ('\n', error_log);
-            fclose (error_log);
-        }
-    }
-}
-
 //////////////////////////////////////////////////////////////////////////////
 // Function to intercept ^C or ^Break typed at the console.                 //
 //////////////////////////////////////////////////////////////////////////////
@@ -583,14 +547,15 @@ void finish_line (void)
     HANDLE hConIn = GetStdHandle (STD_ERROR_HANDLE);
     CONSOLE_SCREEN_BUFFER_INFO coninfo;
 
-    if (hConIn && GetConsoleScreenBufferInfo (hConIn, &coninfo)) {
-        unsigned char spaces = coninfo.dwSize.X - coninfo.dwCursorPosition.X;
+    if (hConIn && GetConsoleScreenBufferInfo (hConIn, &coninfo) &&
+        (coninfo.dwCursorPosition.X || coninfo.dwCursorPosition.Y)) {
+            unsigned char spaces = coninfo.dwSize.X - coninfo.dwCursorPosition.X;
 
-        while (spaces--)
-            fputc (' ', stderr);
+            while (spaces--)
+                fputc (' ', stderr);
     }
     else
-        fputc ('\n', stderr);
+        fprintf (stderr, "                                \n");
 
     fflush (stderr);
 }
@@ -602,7 +567,7 @@ void finish_line (void)
 
 void finish_line (void)
 {
-    fprintf (stderr, "        \n");
+    fprintf (stderr, "                                \n");
     fflush (stderr);
 }
 
