@@ -21,14 +21,14 @@
 #endif
 #endif
 
-#if defined(WIN32)
+#if defined(_WIN32)
 #undef VERSION_OS
 #ifdef _WIN64
 #define VERSION_OS "Win64"
 #else
 #define VERSION_OS "Win32"
 #endif
-#define PACKAGE_VERSION "4.75.0"
+#define PACKAGE_VERSION "4.80.0"
 #endif
 
 #define FALSE 0
