@@ -1,7 +1,7 @@
 ////////////////////////////////////////////////////////////////////////////
 //                           **** WAVPACK ****                            //
 //                  Hybrid Lossless Wavefile Compressor                   //
-//                Copyright (c) 1998 - 2015 David Bryant.                 //
+//                Copyright (c) 1998 - 2016 David Bryant.                 //
 //                          All Rights Reserved.                          //
 //      Distributed under the BSD Software License (see license.txt)      //
 ////////////////////////////////////////////////////////////////////////////
@@ -51,11 +51,18 @@
 #include "wavpack.h"
 #include "utils.h"
 
+#ifdef _WIN32
+#include "win32_unicode_support.h"
+#define fputs fputs_utf8
+#define fprintf fprintf_utf8
+#define fopen(f,m) fopen_utf8(f,m)
+#endif
+
 ///////////////////////////// local variable storage //////////////////////////
 
 static const char *sign_on = "\n"
 " WVGAIN  ReplayGain Scanner/Tagger for WavPack  %s Version %s\n"
-" Copyright (c) 2005 - 2015 David Bryant.  All Rights Reserved.\n\n";
+" Copyright (c) 2005 - 2016 David Bryant.  All Rights Reserved.\n\n";
 
 static const char *version_warning = "\n"
 " WARNING: WVGAIN using libwavpack version %s, expected %s (see README)\n\n";
@@ -105,15 +112,24 @@ static int show_file_info (char *infilename, FILE *dst);
 static float calc_replaygain (uint32_t *histogram);
 static void display_progress (double file_progress);
 
+#ifdef _WIN32
+static void TextToUTF8 (void *string, int len);
+#endif
+
 #define WAVPACK_NO_ERROR    0
 #define WAVPACK_SOFT_ERROR  1
 #define WAVPACK_HARD_ERROR  2
 
-///////////////////////////////////////////////////////////////////////////////////
-// The "main" function for the command-line WavPack ReplayGain Scanner/Processor //
-///////////////////////////////////////////////////////////////////////////////////
+// The "main" function for the command-line WavPack ReplayGain Scanner/Processor.
+// Note that on Windows this is actually a static function that is called from the
+// "real" main() defined immediately afterward that converts the wchar argument list
+// into UTF-8 strings and sets the console to UTF-8 for better Unicode support.
 
-int main (argc, argv) int argc; char **argv;
+#ifdef _WIN32
+static int wvgain_main(int argc, char **argv)
+#else
+int main(int argc, char **argv)
+#endif
 {
 #ifdef __EMX__ /* OS/2 */
     _wildcard (&argc, &argv);
@@ -123,7 +139,6 @@ int main (argc, argv) int argc; char **argv;
     int result = WAVPACK_NO_ERROR;
 
 #if defined(_WIN32)
-    struct _finddata_t _finddata_t;
     char selfname [MAX_PATH];
 
     if (GetModuleFileName (NULL, selfname, sizeof (selfname)) && filespec_name (selfname) &&
@@ -267,8 +282,9 @@ int main (argc, argv) int argc; char **argv;
         // frontends, but could be used for other purposes.
 
         if (*infilename == '@') {
-            FILE *list = fopen (infilename+1, "rt");
-            int di, c;
+            FILE *list = fopen (infilename+1, "rb");
+            char *listbuff = NULL, *cp;
+            int listbytes = 0, di, c;
 
             for (di = file_index; di < num_files - 1; di++)
                 matches [di] = matches [di + 1];
@@ -282,21 +298,37 @@ int main (argc, argv) int argc; char **argv;
                 return 1;
             }
 
-            while ((c = getc (list)) != EOF) {
+            while (1) {
+                int bytes_read;
 
-                while (c == '\n')
-                    c = getc (list);
+                listbuff = realloc (listbuff, listbytes + 1024);
+                memset (listbuff + listbytes, 0, 1024);
+                listbytes += bytes_read = (int) fread (listbuff + listbytes, 1, 1024, list);
 
-                if (c != EOF) {
+                if (bytes_read < 1024)
+                    break;
+            }
+
+#if defined (_WIN32)
+            listbuff = realloc (listbuff, listbytes *= 2);
+            TextToUTF8 (listbuff, listbytes);
+#endif
+            cp = listbuff;
+
+            while ((c = *cp++)) {
+
+                while (c == '\n' || c == '\r')
+                    c = *cp++;
+
+                if (c) {
                     char *fname = malloc (PATH_MAX);
                     int ci = 0;
 
                     do
                         fname [ci++] = c;
-                    while ((c = getc (list)) != '\n' && c != EOF && ci < PATH_MAX);
+                    while ((c = *cp++) != '\n' && c != '\r' && c && ci < PATH_MAX);
 
                     fname [ci++] = '\0';
-                    fname = realloc (fname, ci);
                     matches = realloc (matches, ++num_files * sizeof (*matches));
 
                     for (di = num_files - 1; di > file_index + 1; di--)
@@ -304,14 +336,19 @@ int main (argc, argv) int argc; char **argv;
 
                     matches [++file_index] = fname;
                 }
+
+                if (!c)
+                    break;
             }
 
             fclose (list);
+            free (listbuff);
             free (infilename);
         }
 #if defined (_WIN32)
         else if (filespec_wild (infilename)) {
-            FILE *list = fopen (infilename+1, "rt");
+            wchar_t *winfilename = utf8_to_utf16(infilename);
+            struct _wfinddata_t _wfinddata_t;
             intptr_t file;
             int di;
 
@@ -321,24 +358,28 @@ int main (argc, argv) int argc; char **argv;
             file_index--;
             num_files--;
 
-            if ((file = _findfirst (infilename, &_finddata_t)) != (intptr_t) -1) {
+            if ((file = _wfindfirst (winfilename, &_wfinddata_t)) != (intptr_t) -1) {
                 do {
-                    if (!(_finddata_t.attrib & _A_SUBDIR)) {
+                    char *name_utf8;
+
+                    if (!(_wfinddata_t.attrib & _A_SUBDIR) && (name_utf8 = utf16_to_utf8(_wfinddata_t.name))) {
                         matches = realloc (matches, ++num_files * sizeof (*matches));
 
                         for (di = num_files - 1; di > file_index + 1; di--)
                             matches [di] = matches [di - 1];
 
-                        matches [++file_index] = malloc (strlen (infilename) + strlen (_finddata_t.name) + 10);
+                        matches [++file_index] = malloc (strlen (infilename) + strlen (name_utf8) + 10);
                         strcpy (matches [file_index], infilename);
                         *filespec_name (matches [file_index]) = '\0';
-                        strcat (matches [file_index], _finddata_t.name);
+                        strcat (matches [file_index], name_utf8);
+                        free (name_utf8);
                     }
-                } while (_findnext (file, &_finddata_t) == 0);
+                } while (_wfindnext (file, &_wfinddata_t) == 0);
 
                 _findclose (file);
             }
 
+            free (winfilename);
             free (infilename);
         }
 #endif
@@ -370,9 +411,11 @@ int main (argc, argv) int argc; char **argv;
             if (new_mode) {
                 WavpackContext *wpc;
                 char error [80];
-
+#ifdef _WIN32
+                wpc = WavpackOpenFileInput (matches [file_index], error, OPEN_TAGS | OPEN_FILE_UTF8, 0);
+#else
                 wpc = WavpackOpenFileInput (matches [file_index], error, OPEN_TAGS, 0);
-
+#endif
                 if (wpc) {
                     int alreadyHasTag = WavpackGetTagItem (wpc, album_mode ? "replaygain_album_gain" : "replaygain_track_gain", NULL, 0);
                     WavpackCloseFile (wpc);
@@ -493,6 +536,40 @@ int main (argc, argv) int argc; char **argv;
     return error_count ? 1 : 0;
 }
 
+#ifdef _WIN32
+
+// On Windows, this "real" main() acts as a shell to our static wvgain_main().
+// Its purpose is to convert the wchar command-line arguments into UTF-8 encoded
+// strings and set the console output to UTF-8.
+
+int main(int argc, char **argv)
+{
+    int ret = -1, argc_utf8 = -1;
+    char **argv_utf8 = NULL;
+    char **argv_copy = NULL;
+
+    init_commandline_arguments_utf8(&argc_utf8, &argv_utf8);
+
+    // we have to make a copy of the argv pointer array because the command parser
+    // sometimes modifies them, which is problematic when it comes time to free them
+
+    if (argc_utf8 && argv_utf8) {
+        argv_copy = malloc (sizeof (char*) * argc_utf8);
+        memcpy (argv_copy, argv_utf8, sizeof (char*) * argc_utf8);
+    }
+
+    ret = wvgain_main(argc_utf8, argv_copy);
+
+    if (argv_copy)
+        free (argv_copy);
+
+    free_commandline_arguments_utf8(&argc_utf8, &argv_utf8);
+
+    return ret;
+}
+
+#endif
+
 // Unpack the specified WavPack input file and analyze it for ReplayGain
 // information.
 
@@ -516,6 +593,10 @@ static int analyze_file (char *infilename, uint32_t *histogram, float *peak)
 
     // use library to open WavPack file
 
+#ifdef _WIN32
+    open_flags |= OPEN_FILE_UTF8;
+#endif
+
     if (!ignore_wvc)
         open_flags |= OPEN_WVC;
 
@@ -656,7 +737,11 @@ static int update_file (char *infilename, float track_gain, float track_peak, fl
 
     // use library to open WavPack file
 
+#ifdef _WIN32
+    wpc = WavpackOpenFileInput (infilename, error, OPEN_EDIT_TAGS | OPEN_FILE_UTF8, 0);
+#else
     wpc = WavpackOpenFileInput (infilename, error, OPEN_EDIT_TAGS, 0);
+#endif
 
     if (!wpc) {
         error_line (error);
@@ -757,7 +842,11 @@ static int show_file_info (char *infilename, FILE *dst)
 
     // use library to open WavPack file
 
+#ifdef _WIN32
+    wpc = WavpackOpenFileInput (infilename, error, OPEN_TAGS | OPEN_FILE_UTF8, 0);
+#else
     wpc = WavpackOpenFileInput (infilename, error, OPEN_TAGS, 0);
+#endif
 
     if (!wpc) {
         error_line (error);
@@ -1183,6 +1272,66 @@ static double calc_stereo_rms (float *samples, uint32_t samcnt)
     return 10 * log10 (sum / samcnt) + 90.0 - 3.0;
 }
 
+#ifdef _WIN32
+
+// Convert the Unicode wide-format string into a UTF-8 string using no more
+// than the specified buffer length. The wide-format string must be NULL
+// terminated and the resulting string will be NULL terminated. The actual
+// number of characters converted (not counting terminator) is returned, which
+// may be less than the number of characters in the wide string if the buffer
+// length is exceeded.
+
+static int WideCharToUTF8 (const wchar_t *Wide, unsigned char *pUTF8, int len)
+{
+    const wchar_t *pWide = Wide;
+    int outndx = 0;
+
+    while (*pWide) {
+        if (*pWide < 0x80 && outndx + 1 < len)
+            pUTF8 [outndx++] = (unsigned char) *pWide++;
+        else if (*pWide < 0x800 && outndx + 2 < len) {
+            pUTF8 [outndx++] = (unsigned char) (0xc0 | ((*pWide >> 6) & 0x1f));
+            pUTF8 [outndx++] = (unsigned char) (0x80 | (*pWide++ & 0x3f));
+        }
+        else if (outndx + 3 < len) {
+            pUTF8 [outndx++] = (unsigned char) (0xe0 | ((*pWide >> 12) & 0xf));
+            pUTF8 [outndx++] = (unsigned char) (0x80 | ((*pWide >> 6) & 0x3f));
+            pUTF8 [outndx++] = (unsigned char) (0x80 | (*pWide++ & 0x3f));
+        }
+        else
+            break;
+    }
+
+    pUTF8 [outndx] = 0;
+    return (int)(pWide - Wide);
+}
+
+// Convert a text string into its Unicode UTF-8 format equivalent. The
+// conversion is done in-place so the maximum length of the string buffer must
+// be specified because the string may become longer or shorter. If the
+// resulting string will not fit in the specified buffer size then it is
+// truncated.
+
+static void TextToUTF8 (void *string, int len)
+{
+    if (* (wchar_t *) string == 0xFEFF) {
+        wchar_t *temp = _wcsdup (string);
+
+        WideCharToUTF8 (temp + 1, (unsigned char *) string, len);
+        free (temp);
+    }
+    else {
+        int max_chars = (int) strlen (string);
+        wchar_t *temp = (wchar_t *) malloc ((max_chars + 1) * 2);
+
+        MultiByteToWideChar (CP_ACP, 0, string, -1, temp, max_chars + 1);
+        WideCharToUTF8 (temp, (unsigned char *) string, len);
+        free (temp);
+    }
+}
+
+#endif
+
 //////////////////////////////////////////////////////////////////////////////
 // This function displays the progress status on the title bar of the DOS   //
 // window that WavPack is running in. The "file_progress" argument is for   //
