@@ -1,7 +1,7 @@
 ////////////////////////////////////////////////////////////////////////////
 //                           **** WAVPACK ****                            //
 //                  Hybrid Lossless Wavefile Compressor                   //
-//                Copyright (c) 1998 - 2015 David Bryant.                 //
+//                Copyright (c) 1998 - 2016 David Bryant.                 //
 //                          All Rights Reserved.                          //
 //      Distributed under the BSD Software License (see license.txt)      //
 ////////////////////////////////////////////////////////////////////////////
@@ -13,6 +13,7 @@
 #if defined(_WIN32)
 #define WIN32_LEAN_AND_MEAN
 #include <windows.h>
+#include <conio.h>
 #include <io.h>
 #else
 #if defined(__OS2__)
@@ -48,15 +49,20 @@
 #include "md5.h"
 
 #ifdef _WIN32
+#include "win32_unicode_support.h"
+#define fputs fputs_utf8
+#define fprintf fprintf_utf8
+#define remove(f) unlink_utf8(f)
+#define rename(o,n) rename_utf8(o,n)
+#define fopen(f,m) fopen_utf8(f,m)
 #define strdup(x) _strdup(x)
-#define fileno _fileno
 #endif
 
 ///////////////////////////// local variable storage //////////////////////////
 
 static const char *sign_on = "\n"
 " WVUNPACK  Hybrid Lossless Audio Decompressor  %s Version %s\n"
-" Copyright (c) 1998 - 2015 David Bryant.  All Rights Reserved.\n\n";
+" Copyright (c) 1998 - 2016 David Bryant.  All Rights Reserved.\n\n";
 
 static const char *version_warning = "\n"
 " WARNING: WVUNPACK using libwavpack version %s, expected %s (see README)\n\n";
@@ -83,8 +89,11 @@ static const char *usage =
 #endif
 "          -m  = calculate and display MD5 signature; verify if lossless\n"
 "          -n  = no audio decoding (use with -xx to extract tags only)\n"
+#ifdef _WIN32
+"          --no-utf8-convert = leave tag items in UTF-8 when extracting to files\n"
+"          --pause = pause before exiting (if console window disappears)\n"
+#else
 "          --no-utf8-convert = leave tag items in UTF-8 on extract or display\n"
-#if !defined (_WIN32)
 "          -o FILENAME | PATH = specify output filename or path\n"
 #endif
 "          -q  = quiet (keep console output to a minimum)\n"
@@ -118,7 +127,7 @@ static const char *usage =
 
 int debug_logging_mode;
 
-static char overwrite_all, delete_source, raw_decode, no_utf8_convert, no_audio_decode, file_info,
+static char overwrite_all, delete_source, raw_decode, no_utf8_convert, no_audio_decode, file_info, pause_mode,
     summary, ignore_wvc, quiet_mode, calc_md5, copy_time, blind_decode, wav_decode, set_console_title;
 
 static int num_files, file_index, outbuf_k;
@@ -139,15 +148,24 @@ static void parse_sample_time_index (struct sample_time_index *dst, char *src);
 static int unpack_file (char *infilename, char *outfilename);
 static void display_progress (double file_progress);
 
+#ifdef _WIN32
+static void TextToUTF8 (void *string, int len);
+#endif
+
 #define WAVPACK_NO_ERROR    0
 #define WAVPACK_SOFT_ERROR  1
 #define WAVPACK_HARD_ERROR  2
 
-//////////////////////////////////////////////////////////////////////////////
-// The "main" function for the command-line WavPack decompressor.           //
-//////////////////////////////////////////////////////////////////////////////
+// The "main" function for the command-line WavPack decompressor. Note that on Windows
+// this is actually a static function that is called from the "real" main() defined
+// immediately afterward that converts the wchar argument list into UTF-8 strings
+// and sets the console to UTF-8 for better Unicode support.
 
-int main (argc, argv) int argc; char **argv;
+#ifdef _WIN32
+static int wvunpack_main(int argc, char **argv)
+#else
+int main(int argc, char **argv)
+#endif
 {
 #ifdef __EMX__ /* OS/2 */
     _wildcard (&argc, &argv);
@@ -157,7 +175,6 @@ int main (argc, argv) int argc; char **argv;
     int result;
 
 #if defined(_WIN32)
-    struct _finddata_t _finddata_t;
     char selfname [MAX_PATH];
 
     if (GetModuleFileName (NULL, selfname, sizeof (selfname)) && filespec_name (selfname) &&
@@ -184,7 +201,7 @@ int main (argc, argv) int argc; char **argv;
 #endif
 
 #if defined (_WIN32)
-   set_console_title = 1;      // on Windows, we default to messing with the console title
+    set_console_title = 1;      // on Windows, we default to messing with the console title
 #endif                          // on Linux, this is considered uncool to do by default
 
     // loop through command-line arguments
@@ -206,6 +223,10 @@ int main (argc, argv) int argc; char **argv;
                 printf ("libwavpack %s\n", WavpackGetLibraryVersionString ());
                 return 0;
             }
+#ifdef _WIN32
+            else if (!strcmp (long_option, "pause"))                    // --pause
+                pause_mode = 1;
+#endif
             else if (!strcmp (long_option, "no-utf8-convert"))          // --no-utf8-convert
                 no_utf8_convert = 1;
             else if (!strncmp (long_option, "skip", 4)) {               // --skip
@@ -465,8 +486,9 @@ int main (argc, argv) int argc; char **argv;
         // frontends, but could be used for other purposes.
 
         if (*infilename == '@') {
-            FILE *list = fopen (infilename+1, "rt");
-            int di, c;
+            FILE *list = fopen (infilename+1, "rb");
+            char *listbuff = NULL, *cp;
+            int listbytes = 0, di, c;
 
             for (di = file_index; di < num_files - 1; di++)
                 matches [di] = matches [di + 1];
@@ -480,21 +502,37 @@ int main (argc, argv) int argc; char **argv;
                 return 1;
             }
 
-            while ((c = getc (list)) != EOF) {
+            while (1) {
+                int bytes_read;
 
-                while (c == '\n')
-                    c = getc (list);
+                listbuff = realloc (listbuff, listbytes + 1024);
+                memset (listbuff + listbytes, 0, 1024);
+                listbytes += bytes_read = (int) fread (listbuff + listbytes, 1, 1024, list);
 
-                if (c != EOF) {
+                if (bytes_read < 1024)
+                    break;
+            }
+
+#if defined (_WIN32)
+            listbuff = realloc (listbuff, listbytes *= 2);
+            TextToUTF8 (listbuff, listbytes);
+#endif
+            cp = listbuff;
+
+            while ((c = *cp++)) {
+
+                while (c == '\n' || c == '\r')
+                    c = *cp++;
+
+                if (c) {
                     char *fname = malloc (PATH_MAX);
                     int ci = 0;
 
                     do
                         fname [ci++] = c;
-                    while ((c = getc (list)) != '\n' && c != EOF && ci < PATH_MAX);
+                    while ((c = *cp++) != '\n' && c != '\r' && c && ci < PATH_MAX);
 
                     fname [ci++] = '\0';
-                    fname = realloc (fname, ci);
                     matches = realloc (matches, ++num_files * sizeof (*matches));
 
                     for (di = num_files - 1; di > file_index + 1; di--)
@@ -502,14 +540,19 @@ int main (argc, argv) int argc; char **argv;
 
                     matches [++file_index] = fname;
                 }
+
+                if (!c)
+                    break;
             }
 
             fclose (list);
+            free (listbuff);
             free (infilename);
         }
 #if defined (_WIN32)
         else if (filespec_wild (infilename)) {
-            FILE *list = fopen (infilename+1, "rt");
+            wchar_t *winfilename = utf8_to_utf16(infilename);
+            struct _wfinddata_t _wfinddata_t;
             intptr_t file;
             int di;
 
@@ -519,24 +562,28 @@ int main (argc, argv) int argc; char **argv;
             file_index--;
             num_files--;
 
-            if ((file = _findfirst (infilename, &_finddata_t)) != (intptr_t) -1) {
+            if ((file = _wfindfirst (winfilename, &_wfinddata_t)) != (intptr_t) -1) {
                 do {
-                    if (!(_finddata_t.attrib & _A_SUBDIR)) {
+                    char *name_utf8;
+
+                    if (!(_wfinddata_t.attrib & _A_SUBDIR) && (name_utf8 = utf16_to_utf8(_wfinddata_t.name))) {
                         matches = realloc (matches, ++num_files * sizeof (*matches));
 
                         for (di = num_files - 1; di > file_index + 1; di--)
                             matches [di] = matches [di - 1];
 
-                        matches [++file_index] = malloc (strlen (infilename) + strlen (_finddata_t.name) + 10);
+                        matches [++file_index] = malloc (strlen (infilename) + strlen (name_utf8) + 10);
                         strcpy (matches [file_index], infilename);
                         *filespec_name (matches [file_index]) = '\0';
-                        strcat (matches [file_index], _finddata_t.name);
+                        strcat (matches [file_index], name_utf8);
+                        free (name_utf8);
                     }
-                } while (_findnext (file, &_finddata_t) == 0);
+                } while (_wfindnext (file, &_wfinddata_t) == 0);
 
                 _findclose (file);
             }
 
+            free (winfilename);
             free (infilename);
         }
 #endif
@@ -548,7 +595,8 @@ int main (argc, argv) int argc; char **argv;
     // be passed on the command-line, but could be used for other purposes.
 
     if (outfilename && outfilename [0] == '@') {
-        FILE *list = fopen (outfilename+1, "rt");
+        char listbuff [PATH_MAX * 2], *lp = listbuff;
+        FILE *list = fopen (outfilename+1, "rb");
         int c;
 
         if (list == NULL) {
@@ -557,14 +605,21 @@ int main (argc, argv) int argc; char **argv;
             return 1;
         }
 
-        while ((c = getc (list)) == '\n');
+        memset (listbuff, 0, sizeof (listbuff));
+        c = (int) fread (listbuff, 1, sizeof (listbuff) - 1, list);   // assign c only to suppress warning
 
-        if (c != EOF) {
+#if defined (_WIN32)
+        TextToUTF8 (listbuff, PATH_MAX * 2);
+#endif
+
+        while ((c = *lp++) == '\n' || c == '\r');
+
+        if (c) {
             int ci = 0;
 
             do
                 outfilename [ci++] = c;
-            while ((c = getc (list)) != '\n' && c != EOF && ci < PATH_MAX);
+            while ((c = *lp++) != '\n' && c != '\r' && c && ci < PATH_MAX);
 
             outfilename [ci] = '\0';
         }
@@ -672,6 +727,48 @@ int main (argc, argv) int argc; char **argv;
     return error_count ? 1 : 0;
 }
 
+#ifdef _WIN32
+
+// On Windows, this "real" main() acts as a shell to our static wvunpack_main().
+// Its purpose is to convert the wchar command-line arguments into UTF-8 encoded
+// strings.
+
+int main(int argc, char **argv)
+{
+    int ret = -1, argc_utf8 = -1;
+    char **argv_utf8 = NULL;
+    char **argv_copy = NULL;
+
+    init_commandline_arguments_utf8(&argc_utf8, &argv_utf8);
+
+    // we have to make a copy of the argv pointer array because the command parser
+    // sometimes modifies them, which is problematic when it comes time to free them
+
+    if (argc_utf8 && argv_utf8) {
+        argv_copy = malloc (sizeof (char*) * argc_utf8);
+        memcpy (argv_copy, argv_utf8, sizeof (char*) * argc_utf8);
+    }
+
+    ret = wvunpack_main(argc_utf8, argv_copy);
+
+    if (argv_copy)
+        free (argv_copy);
+
+    free_commandline_arguments_utf8(&argc_utf8, &argv_utf8);
+
+    if (pause_mode) {
+        fprintf (stderr, "\nPress any key to continue . . . ");
+        fflush (stderr);
+        while (!_kbhit ());
+        _getch ();
+        fprintf (stderr, "\n");
+    }
+
+    return ret;
+}
+
+#endif
+
 // Parse the parameter of the --skip and --until commands, which are of the form:
 //   [+|-] [samples | hh:mm:ss.ss]
 // The value is returned in a double (in the "dst" struct) as either samples or
@@ -728,7 +825,7 @@ static FILE *open_output_file (char *filename, char **tempfilename)
 
     if (*filename == '-') {
 #if defined(_WIN32)
-        _setmode (fileno (stdout), O_BINARY);
+        _setmode (_fileno (stdout), O_BINARY);
 #endif
 #if defined(__OS2__)
         setmode (fileno (stdout), O_BINARY);
@@ -819,6 +916,8 @@ static int write_riff_header (FILE *outfile, WavpackContext *wpc, uint32_t total
 static int dump_tag_item_to_file (WavpackContext *wpc, const char *tag_item, FILE *dst, char *fn);
 static void dump_file_info (WavpackContext *wpc, char *name, FILE *dst);
 
+#define TEMP_BUFFER_SAMPLES 4096L   // composite samples in temporary buffer used during unpacking
+
 static int unpack_file (char *infilename, char *outfilename)
 {
     int result = WAVPACK_NO_ERROR, md5_diff = FALSE, created_riff_header = FALSE;
@@ -843,6 +942,10 @@ static int unpack_file (char *infilename, char *outfilename)
 
     // use library to open WavPack file
 
+#ifdef _WIN32
+    open_flags |= OPEN_FILE_UTF8;
+#endif
+
     if ((outfilename && !raw_decode && !blind_decode && !wav_decode &&
         !skip.value_is_valid && !until.value_is_valid) || summary > 1)
             open_flags |= OPEN_WRAPPER;
@@ -1021,7 +1124,7 @@ static int unpack_file (char *infilename, char *outfilename)
             created_riff_header = TRUE;
     }
 
-    temp_buffer = malloc (4096L * num_channels * 4);
+    temp_buffer = malloc (TEMP_BUFFER_SAMPLES * num_channels * sizeof (temp_buffer [0]));
 
     while (result == WAVPACK_NO_ERROR) {
         uint32_t samples_to_unpack, samples_unpacked;
@@ -1029,11 +1132,11 @@ static int unpack_file (char *infilename, char *outfilename)
         if (output_buffer) {
             samples_to_unpack = (output_buffer_size - (uint32_t)(output_pointer - output_buffer)) / bytes_per_sample;
 
-            if (samples_to_unpack > 4096)
-                samples_to_unpack = 4096;
+            if (samples_to_unpack > TEMP_BUFFER_SAMPLES)
+                samples_to_unpack = TEMP_BUFFER_SAMPLES;
         }
         else
-            samples_to_unpack = 4096;
+            samples_to_unpack = TEMP_BUFFER_SAMPLES;
 
         if (until_samples_total && samples_to_unpack > until_samples_total - total_unpacked_samples)
             samples_to_unpack = until_samples_total - total_unpacked_samples;
@@ -1852,7 +1955,7 @@ static int dump_tag_item_to_file (WavpackContext *wpc, const char *tag_item, FIL
                 return value_len;
 
 #if defined(_WIN32)
-            _setmode (fileno (dst), O_TEXT);
+            _setmode (_fileno (dst), O_TEXT);
 #endif
 #if defined(__OS2__)
             setmode (fileno (dst), O_TEXT);
@@ -1876,7 +1979,7 @@ static int dump_tag_item_to_file (WavpackContext *wpc, const char *tag_item, FIL
 
                     if (dst) {
 #if defined(_WIN32)
-                        _setmode (fileno (dst), O_BINARY);
+                        _setmode (_fileno (dst), O_BINARY);
 #endif
 #if defined(__OS2__)
                         setmode (fileno (dst), O_BINARY);
@@ -1948,7 +2051,11 @@ static void dump_UTF8_string (char *string, FILE *dst)
 
         *p = 0;
 
+#ifdef _WIN32
+        if (!no_utf8_convert && dst != stdout && dst != stderr)
+#else
         if (!no_utf8_convert)
+#endif
             UTF8ToAnsi (temp, len * 2);
 
         fputs (temp, dst);
@@ -1963,7 +2070,7 @@ static void dump_UTF8_string (char *string, FILE *dst)
 // and will also be NULL terminated. The number of characters converted will
 // be returned (not counting terminator).
 
-static int UTF8ToWideChar (const unsigned char *pUTF8, unsigned short *pWide)
+static int UTF8ToWideChar (const unsigned char *pUTF8, wchar_t *pWide)
 {
     int trail_bytes = 0;
     int chrcnt = 0;
@@ -2003,6 +2110,62 @@ static int UTF8ToWideChar (const unsigned char *pUTF8, unsigned short *pWide)
     return chrcnt;
 }
 
+// Convert the Unicode wide-format string into a UTF-8 string using no more
+// than the specified buffer length. The wide-format string must be NULL
+// terminated and the resulting string will be NULL terminated. The actual
+// number of characters converted (not counting terminator) is returned, which
+// may be less than the number of characters in the wide string if the buffer
+// length is exceeded.
+
+static int WideCharToUTF8 (const wchar_t *Wide, unsigned char *pUTF8, int len)
+{
+    const wchar_t *pWide = Wide;
+    int outndx = 0;
+
+    while (*pWide) {
+        if (*pWide < 0x80 && outndx + 1 < len)
+            pUTF8 [outndx++] = (unsigned char) *pWide++;
+        else if (*pWide < 0x800 && outndx + 2 < len) {
+            pUTF8 [outndx++] = (unsigned char) (0xc0 | ((*pWide >> 6) & 0x1f));
+            pUTF8 [outndx++] = (unsigned char) (0x80 | (*pWide++ & 0x3f));
+        }
+        else if (outndx + 3 < len) {
+            pUTF8 [outndx++] = (unsigned char) (0xe0 | ((*pWide >> 12) & 0xf));
+            pUTF8 [outndx++] = (unsigned char) (0x80 | ((*pWide >> 6) & 0x3f));
+            pUTF8 [outndx++] = (unsigned char) (0x80 | (*pWide++ & 0x3f));
+        }
+        else
+            break;
+    }
+
+    pUTF8 [outndx] = 0;
+    return (int)(pWide - Wide);
+}
+
+// Convert a text string into its Unicode UTF-8 format equivalent. The
+// conversion is done in-place so the maximum length of the string buffer must
+// be specified because the string may become longer or shorter. If the
+// resulting string will not fit in the specified buffer size then it is
+// truncated.
+
+static void TextToUTF8 (void *string, int len)
+{
+    if (* (wchar_t *) string == 0xFEFF) {
+        wchar_t *temp = _wcsdup (string);
+
+        WideCharToUTF8 (temp + 1, (unsigned char *) string, len);
+        free (temp);
+    }
+    else {
+        int max_chars = (int) strlen (string);
+        wchar_t *temp = (wchar_t *) malloc ((max_chars + 1) * 2);
+
+        MultiByteToWideChar (CP_ACP, 0, string, -1, temp, max_chars + 1);
+        WideCharToUTF8 (temp, (unsigned char *) string, len);
+        free (temp);
+    }
+}
+
 #endif
 
 // Convert a Unicode UTF-8 format string into its Ansi equivalent. The
@@ -2015,7 +2178,7 @@ static void UTF8ToAnsi (char *string, int len)
 {
     int max_chars = (int) strlen (string);
 #if defined (_WIN32)
-    unsigned short *temp = malloc ((max_chars + 1) * 2);
+    wchar_t *temp = malloc ((max_chars + 1) * 2);
     int act_chars = UTF8ToWideChar (string, temp);
 
     while (act_chars) {
