@@ -9,7 +9,7 @@
 // open_utils.c
 
 // This module provides all the code required to open an existing WavPack file
-// for reading, either by filename or by using a reader callback mechanism. This
+// for reading by using a reader callback mechanism (NOT a filename). This
 // includes the code required to find and parse WavPack blocks, process any
 // included metadata, and queue up the bitstreams containing the encoded audio
 // data. It does not the actual code to unpack audio data and this was done so
@@ -21,155 +21,6 @@
 
 #include "wavpack_local.h"
 
-// This code provides an interface between the reader callback mechanism that
-// WavPack uses internally and the standard fstream C library. This allows an
-// "open" call for WavPack files that accepts a filename (which is located at
-// the end of the #ifdef).
-
-#ifndef NO_USE_FSTREAMS
-
-#include <fcntl.h>
-#include <sys/stat.h>
-
-#if defined (_WIN32) || defined (__OS2__)
-#include <io.h>
-#endif
-
-#ifdef _WIN32
-#define fileno _fileno
-#endif
-
-static int32_t read_bytes (void *id, void *data, int32_t bcount)
-{
-    return (int32_t) fread (data, 1, bcount, (FILE*) id);
-}
-
-static uint32_t get_pos (void *id)
-{
-    return ftell ((FILE*) id);
-}
-
-static int set_pos_abs (void *id, uint32_t pos)
-{
-    return fseek (id, pos, SEEK_SET);
-}
-
-static int set_pos_rel (void *id, int32_t delta, int mode)
-{
-    return fseek (id, delta, mode);
-}
-
-static int push_back_byte (void *id, int c)
-{
-    return ungetc (c, id);
-}
-
-static uint32_t get_length (void *id)
-{
-    FILE *file = id;
-    struct stat statbuf;
-
-    if (!file || fstat (fileno (file), &statbuf) || !(statbuf.st_mode & S_IFREG))
-        return 0;
-
-    return statbuf.st_size;
-}
-
-static int can_seek (void *id)
-{
-    FILE *file = id;
-    struct stat statbuf;
-
-    return file && !fstat (fileno (file), &statbuf) && (statbuf.st_mode & S_IFREG);
-}
-
-static int32_t write_bytes (void *id, void *data, int32_t bcount)
-{
-    return (int32_t) fwrite (data, 1, bcount, (FILE*) id);
-}
-
-static WavpackStreamReader freader = {
-    read_bytes, get_pos, set_pos_abs, set_pos_rel, push_back_byte, get_length, can_seek,
-    write_bytes
-};
-
-// This function attempts to open the specified WavPack file for reading. If
-// this fails for any reason then an appropriate message is copied to "error"
-// (which must accept 80 characters) and NULL is returned, otherwise a
-// pointer to a WavpackContext structure is returned (which is used to call
-// all other functions in this module). A filename beginning with "-" is
-// assumed to be stdin. The "flags" argument has the following bit mask
-// values to specify details of the open operation:
-
-// OPEN_WVC:  attempt to open/read "correction" file
-// OPEN_TAGS:  attempt to read ID3v1 / APEv2 tags (requires seekable file)
-// OPEN_WRAPPER:  make audio wrapper available (i.e. RIFF) to caller
-// OPEN_2CH_MAX:  open only first stream of multichannel file (usually L/R)
-// OPEN_NORMALIZE:  normalize floating point data to +/- 1.0 (w/ offset exp)
-// OPEN_STREAMING:  blindly unpacks blocks w/o regard to header file position
-// OPEN_EDIT_TAGS:  allow editing of tags (file must be writable)
-
-// Version 4.2 of the WavPack library adds the OPEN_STREAMING flag. This is
-// essentially a "raw" mode where the library will simply decode any blocks
-// fed it through the reader callback, regardless of where those blocks came
-// from in a stream. The only requirement is that complete WavPack blocks are
-// fed to the decoder (and this may require multiple blocks in multichannel
-// mode) and that complete blocks are decoded (even if all samples are not
-// actually required). All the blocks must contain the same number of channels
-// and bit resolution, and the correction data must be either present or not.
-// All other parameters may change from block to block (like lossy/lossless).
-// Obviously, in this mode any seeking must be performed by the application
-// (and again, decoding must start at the beginning of the block containing
-// the seek sample).
-
-WavpackContext *WavpackOpenFileInput (const char *infilename, char *error, int flags, int norm_offset)
-{
-    char *file_mode = (flags & OPEN_EDIT_TAGS) ? "r+b" : "rb";
-    FILE *wv_id, *wvc_id;
-    WavpackContext *wpc;
-
-    if (*infilename == '-') {
-        wv_id = stdin;
-#if defined(_WIN32)
-        _setmode (fileno (stdin), O_BINARY);
-#endif
-#if defined(__OS2__)
-        setmode (fileno (stdin), O_BINARY);
-#endif
-    }
-    else if ((wv_id = fopen (infilename, file_mode)) == NULL) {
-        if (error) strcpy (error, (flags & OPEN_EDIT_TAGS) ? "can't open file for editing" : "can't open file");
-        return NULL;
-    }
-
-    if (wv_id != stdin && (flags & OPEN_WVC)) {
-        char *in2filename = malloc (strlen (infilename) + 10);
-
-        strcpy (in2filename, infilename);
-        strcat (in2filename, "c");
-        wvc_id = fopen (in2filename, "rb");
-        free (in2filename);
-    }
-    else
-        wvc_id = NULL;
-
-    wpc = WavpackOpenFileInputEx (&freader, wv_id, wvc_id, error, flags, norm_offset);
-
-    if (!wpc) {
-        if (wv_id)
-            fclose (wv_id);
-
-        if (wvc_id)
-            fclose (wvc_id);
-    }
-    else
-        wpc->close_files = TRUE;
-
-    return wpc;
-}
-
-#endif
-
 // This function is identical to WavpackOpenFileInput() except that instead
 // of providing a filename to open, the caller provides a pointer to a set of
 // reader callbacks and instances of up to two streams. The first of these
@@ -411,6 +262,7 @@ int unpack_init (WavpackContext *wpc)
     unsigned char *blockptr, *block2ptr;
     WavpackMetadata wpmd;
 
+    wps->num_terms = 0;
     wps->mute_error = FALSE;
     wps->crc = wps->crc_x = 0xffffffff;
     CLEAR (wps->wvbits);
@@ -565,7 +417,7 @@ static int read_float_info (WavpackStream *wps, WavpackMetadata *wpmd)
 
 static int read_channel_info (WavpackContext *wpc, WavpackMetadata *wpmd)
 {
-    int bytecnt = wpmd->byte_length, shift = 0;
+    int bytecnt = wpmd->byte_length, shift = 0, mask_bits;
     unsigned char *byteptr = wpmd->data;
     uint32_t mask = 0;
 
@@ -599,6 +451,10 @@ static int read_channel_info (WavpackContext *wpc, WavpackMetadata *wpmd)
             return FALSE;
 
         wpc->config.channel_mask = mask;
+
+        for (mask_bits = 0; mask; mask >>= 1)
+            if ((mask & 1) && ++mask_bits > wpc->config.num_channels)
+                return FALSE;
     }
 
     return TRUE;
