@@ -80,8 +80,14 @@ int32_t unpack_samples (WavpackContext *wpc, int32_t *buffer, uint32_t sample_co
     struct decorr_pass *dpp;
     int tcount, m = 0;
 
-    if (wps->sample_index + sample_count > wps->wphdr.block_index + wps->wphdr.block_samples)
-        sample_count = wps->wphdr.block_index + wps->wphdr.block_samples - wps->sample_index;
+    // don't attempt to decode past the end of the block, but watch out for overflow!
+
+    if (wps->sample_index + sample_count > wps->wphdr.block_index + wps->wphdr.block_samples &&
+        wps->wphdr.block_index + wps->wphdr.block_samples - wps->sample_index < sample_count)
+            sample_count = wps->wphdr.block_index + wps->wphdr.block_samples - wps->sample_index;
+
+    if (wps->wphdr.block_index > wps->sample_index || wps->wphdr.block_samples < sample_count)
+        wps->mute_error = TRUE;
 
     if (wps->mute_error) {
         if (wpc->reduced_channels == 1 || wpc->config.num_channels == 1 || (flags & MONO_FLAG))
@@ -94,7 +100,7 @@ int32_t unpack_samples (WavpackContext *wpc, int32_t *buffer, uint32_t sample_co
     }
 
     if ((flags & HYBRID_FLAG) && !wps->block2buff)
-        mute_limit *= 2;
+        mute_limit = (mute_limit * 2) + 128;
 
     //////////////// handle lossless or hybrid lossy mono data /////////////////
 
@@ -126,19 +132,14 @@ int32_t unpack_samples (WavpackContext *wpc, int32_t *buffer, uint32_t sample_co
                 DECORR_MONO_PASS_CONT (dpp, buffer + pre_samples, sample_count - pre_samples,
                     ((flags & MAG_MASK) >> MAG_LSB) > 15);
             }
-
-        for (bptr = buffer; bptr < eptr; ++bptr) {
-            if (labs (bptr [0]) > mute_limit) {
-                i = (uint32_t)(bptr - buffer);
-                break;
-            }
-
-            crc = crc * 3 + bptr [0];
-        }
 #else
         for (tcount = wps->num_terms, dpp = wps->decorr_passes; tcount--; dpp++)
             decorr_mono_pass (dpp, buffer, sample_count);
+#endif
 
+#ifndef LOSSY_MUTE
+        if (!(flags & HYBRID_FLAG))
+#endif
         for (bptr = buffer; bptr < eptr; ++bptr) {
             if (labs (bptr [0]) > mute_limit) {
                 i = (uint32_t)(bptr - buffer);
@@ -147,7 +148,6 @@ int32_t unpack_samples (WavpackContext *wpc, int32_t *buffer, uint32_t sample_co
 
             crc = crc * 3 + bptr [0];
         }
-#endif
     }
 
     /////////////// handle lossless or hybrid lossy stereo data ///////////////
@@ -200,6 +200,9 @@ int32_t unpack_samples (WavpackContext *wpc, int32_t *buffer, uint32_t sample_co
             for (bptr = buffer; bptr < eptr; bptr += 2)
                 crc += (crc << 3) + (bptr [0] << 1) + bptr [0] + bptr [1];
 
+#ifndef LOSSY_MUTE
+        if (!(flags & HYBRID_FLAG))
+#endif
         for (bptr = buffer; bptr < eptr; bptr += 16)
             if (labs (bptr [0]) > mute_limit || labs (bptr [1]) > mute_limit) {
                 i = (uint32_t)(bptr - buffer) / 2;
@@ -260,10 +263,9 @@ int32_t unpack_samples (WavpackContext *wpc, int32_t *buffer, uint32_t sample_co
 
             crc += (crc << 1) + read_word;
 
-#ifdef LOSSY_MUTE
             if (labs (read_word) > mute_limit)
                 break;
-#endif
+
             *bptr++ = read_word;
         }
 
@@ -432,10 +434,9 @@ int32_t unpack_samples (WavpackContext *wpc, int32_t *buffer, uint32_t sample_co
                 right = right_c;
             }
 
-#ifdef LOSSY_MUTE
             if (labs (left) > mute_limit || labs (right) > mute_limit)
                 break;
-#endif
+
             crc += (crc << 3) + (left << 1) + left + right;
             *bptr++ = left;
             *bptr++ = right;
@@ -494,9 +495,8 @@ int32_t unpack_samples (WavpackContext *wpc, int32_t *buffer, uint32_t sample_co
 // General function to perform mono decorrelation pass on specified buffer
 // (although since this is the reverse function it might technically be called
 // "correlation" instead). This version handles all sample resolutions and
-// weight deltas. The dpp->samples_X[] data is *not* returned normalized for
-// term values 1-8, so it should be normalized if it is going to be used to
-// call this function again.
+// weight deltas. The dpp->samples_X[] data is returned normalized for term
+// values 1-8.
 
 static void decorr_mono_pass (struct decorr_pass *dpp, int32_t *buffer, int32_t sample_count)
 {
