@@ -11,13 +11,13 @@
 #ifndef WAVPACK_LOCAL_H
 #define WAVPACK_LOCAL_H
 
-#if defined(WIN32)
+#if defined(_WIN32)
 #define FASTCALL __fastcall
 #else
 #define FASTCALL
 #endif
 
-#if defined(WIN32) || \
+#if defined(_WIN32) || \
     (defined(BYTE_ORDER) && defined(LITTLE_ENDIAN) && (BYTE_ORDER == LITTLE_ENDIAN)) || \
     (defined(__BYTE_ORDER__) && defined(__ORDER_LITTLE_ENDIAN__) && (__BYTE_ORDER__ == __ORDER_LITTLE_ENDIAN__))
 #define BITSTREAM_SHORTS    // use "shorts" for reading/writing bitstreams
@@ -54,6 +54,7 @@ typedef __int8  int8_t;
 typedef int32_t f32;
 
 #define get_mantissa(f)     ((f) & 0x7fffff)
+#define get_magnitude(f)    ((f) & 0x7fffffff)
 #define get_exponent(f)     (((f) >> 23) & 0xff)
 #define get_sign(f)         (((f) >> 31) & 0x1)
 
@@ -413,10 +414,11 @@ typedef struct {
     void *wv_out, *wvc_out;
 
     WavpackStreamReader *reader;
+    int (*close_file)(void*);
     void *wv_in, *wvc_in;
 
     uint32_t filelen, file2len, filepos, file2pos, total_samples, crc_errors, first_flags;
-    int wvc_flag, open_flags, norm_offset, reduced_channels, lossy_blocks, close_files;
+    int wvc_flag, open_flags, norm_offset, reduced_channels, lossy_blocks;
     uint32_t block_samples, ave_block_samples, block_boundary, max_samples, acc_samples, initial_index, riff_trailer_bytes;
     int riff_header_added, riff_header_created;
     M_Tag m_tag;
@@ -650,9 +652,9 @@ int32_t nosend_word (WavpackStream *wps, int32_t value, int chan);
 void scan_word (WavpackStream *wps, int32_t *samples, uint32_t num_samples, int dir);
 void update_error_limit (WavpackStream *wps);
 
-const uint32_t bitset [32];
-const uint32_t bitmask [32];
-const char nbits_table [256];
+extern const uint32_t bitset [32];
+extern const uint32_t bitmask [32];
+extern const char nbits_table [256];
 
 int log2s (int32_t value);
 int32_t exp2s (int log);
@@ -691,6 +693,7 @@ WavpackContext *WavpackOpenFileInput (const char *infilename, char *error, int f
 #define OPEN_STREAMING  0x20    // "streaming" mode blindly unpacks blocks
                                 // w/o regard to header file position info
 #define OPEN_EDIT_TAGS  0x40    // allow editing of tags
+#define OPEN_FILE_UTF8  0x80    // assume filenames are UTF-8 encoded, not ANSI (Windows only)
 
 int WavpackGetMode (WavpackContext *wpc);
 
@@ -734,7 +737,7 @@ void *WavpackGetWrapperLocation (void *first_block, uint32_t *size);
 /////////////////////////////////// common utilities ////////////////////////////////////
 // module: common_utils.c
 
-const uint32_t sample_rates [16];
+extern const uint32_t sample_rates [16];
 uint32_t WavpackGetLibraryVersion (void);
 const char *WavpackGetLibraryVersionString (void);
 uint32_t WavpackGetSampleRate (WavpackContext *wpc);
